# frozen_string_literal: true
module Hyrax
  ##
  # Store a file uploaded by a user.
  #
  # Eventually these files get attached to {FileSet}s and pushed into Fedora.
  class UploadedFile < ActiveRecord::Base
    self.table_name = 'uploaded_files'
    mount_uploader :file, UploadedFileUploader
    alias uploader file
    has_many :job_io_wrappers,
             inverse_of: 'uploaded_file',
             class_name: 'JobIoWrapper',
             dependent: :destroy
    belongs_to :user, class_name: '::User'
    
    before_create :set_characterization_data
    ##
    # Associate a {FileSet} with this uploaded file.
    #
    # @param [Hyrax::Resource, ActiveFedora::Base] file_set
    # @return [void]
    def add_file_set!(file_set)
      uri = case file_set
            when ActiveFedora::Base
              file_set.uri
            when Hyrax::Resource
              file_set.id
            end
      update!(file_set_uri: uri)
    end

    private

    def set_characterization_data
      if file.file.content_type.include?('image')
        image = MiniMagick::Image.open(file.path)

        self.characterization_data[:height]  = image[:height]
        self.characterization_data[:width]  = image[:width]
      end

      self.characterization_data[:content_type]  = Marcel::Magic.by_path(file.path).to_s
      self.characterization_data[:file_size]  = file.size
      self.characterization_data[:original_checksum] = Digest::SHA1.file(file.path).to_s
    end
  end
end